@@ -32,6 +32,8 @@ use IPC::Open2;
 use File::Path;
 use File::Basename;
 use Dpkg::Control::Hash;
+use Debian::Dgit::ExitStatus;
+use Debian::Dgit::I18n;
 
 BEGIN {
     use Exporter   ();
@@ -43,6 +45,8 @@ BEGIN {
 		      dep14_version_mangle
                       debiantags debiantag_old debiantag_new
 		      debiantag_maintview
+		      upstreamversion
+		      stripepoch source_file_leafname is_orig_file_of_p_v
 		      server_branch server_ref
                       stat_exists link_ltarget
 		      hashfile
@@ -55,14 +59,18 @@ BEGIN {
                       git_for_each_tag_referring is_fast_fwd
 		      git_check_unmodified
 		      git_reflog_action_msg  git_update_ref_cmd
+		      make_commit_text
+		      reflog_cache_insert reflog_cache_lookup
                       $package_re $component_re $deliberately_re
 		      $distro_re $versiontag_re $series_filename_re
+		      $orig_f_comp_re $orig_f_sig_re $orig_f_tail_re
 		      $extra_orig_namepart_re
 		      $git_null_obj
                       $branchprefix
 		      $ffq_refprefix $gdrlast_refprefix
                       initdebug enabledebug enabledebuglevel
                       printdebug debugcmd
+                      $printdebug_when_debuglevel $debugcmd_when_debuglevel
                       $debugprefix *debuglevel *DEBUG
                       shellquote printcmd messagequote
                       $negate_harmful_gitattrs
@@ -91,9 +99,19 @@ our $versiontag_re = qr{[-+.\%_0-9a-zA-Z/]+};
 our $branchprefix = 'dgit';
 our $series_filename_re = qr{(?:^|\.)series(?!\n)$}s;
 our $extra_orig_namepart_re = qr{[-0-9a-z]+};
+our $orig_f_comp_re = qr{orig(?:-$extra_orig_namepart_re)?};
+our $orig_f_sig_re = '\\.(?:asc|gpg|pgp)';
+our $orig_f_tail_re = "$orig_f_comp_re\\.tar(?:\\.\\w+)?(?:$orig_f_sig_re)?";
 our $git_null_obj = '0' x 40;
 our $ffq_refprefix = 'ffq-prev';
 our $gdrlast_refprefix = 'debrebase-last';
+our $printdebug_when_debuglevel = 1;
+our $debugcmd_when_debuglevel = 1;
+
+# these three all go together, only valid after record_maindir
+our $maindir;
+our $maindir_gitdir;
+our $maindir_gitcommon;
 
 # policy hook exit status bits
 # see dgit-repos-server head comment for documentation
@@ -130,18 +148,18 @@ sub setup_sigwarn () {
 
 sub initdebug ($) { 
     ($debugprefix) = @_;
-    open DEBUG, ">/dev/null" or die $!;
+    open DEBUG, ">/dev/null" or confess $!;
 }
 
 sub enabledebug () {
-    open DEBUG, ">&STDERR" or die $!;
+    open DEBUG, ">&STDERR" or confess $!;
     DEBUG->autoflush(1);
     $debuglevel ||= 1;
 }
     
 sub enabledebuglevel ($) {
     my ($newlevel) = @_; # may be undef (eg from env var)
-    die if $debuglevel;
+    confess if $debuglevel;
     $newlevel //= 0;
     $newlevel += 0;
     return unless $newlevel;
@@ -150,7 +168,21 @@ sub enabledebuglevel ($) {
 }
     
 sub printdebug {
-    print DEBUG $debugprefix, @_ or die $! if $debuglevel>0;
+    # Prints a prefix, and @_, to DEBUG.  @_ should normally contain
+    # a trailing \n.
+
+    # With no (or only empty) arguments just prints the prefix and
+    # leaves the caller to do more with DEBUG.  The caller should make
+    # sure then to call printdebug with something ending in "\n" to
+    # get the prefix right in subsequent calls.
+
+    return unless $debuglevel >= $printdebug_when_debuglevel;
+    our $printdebug_noprefix;
+    print DEBUG $debugprefix unless $printdebug_noprefix;
+    pop @_ while @_ and !length $_[-1];
+    return unless @_;
+    print DEBUG @_ or confess $!;
+    $printdebug_noprefix = $_[-1] !~ m{\n$};
 }
 
 sub messagequote ($) {
@@ -166,7 +198,7 @@ sub messagequote ($) {
 sub shellquote {
     my @out;
     local $_;
-    defined or confess 'internal error' foreach @_;
+    defined or confess __ 'internal error' foreach @_;
     foreach my $a (@_) {
 	$_ = $a;
 	if (!length || m{[^-=_./:0-9a-z]}i) {
@@ -182,14 +214,15 @@ sub shellquote {
 sub printcmd {
     my $fh = shift @_;
     my $intro = shift @_;
-    print $fh $intro," " or die $!;
-    print $fh shellquote @_ or die $!;
-    print $fh "\n" or die $!;
+    print $fh $intro," " or confess $!;
+    print $fh shellquote @_ or confess $!;
+    print $fh "\n" or confess $!;
 }
 
 sub debugcmd {
     my $extraprefix = shift @_;
-    printcmd(\*DEBUG,$debugprefix.$extraprefix,@_) if $debuglevel>0;
+    printcmd(\*DEBUG,$debugprefix.$extraprefix,@_)
+	if $debuglevel >= $debugcmd_when_debuglevel;
 }
 
 sub dep14_version_mangle ($) {
@@ -220,6 +253,30 @@ sub debiantags ($$) {
     map { $_->($version, $distro) } (\&debiantag_new, \&debiantag_old);
 }
 
+sub stripepoch ($) {
+    my ($vsn) = @_;
+    $vsn =~ s/^\d+\://;
+    return $vsn;
+}
+
+sub upstreamversion ($) {
+    my ($vsn) = @_;
+    $vsn =~ s/-[^-]+$//;
+    return $vsn;
+}
+
+sub source_file_leafname ($$$) {
+    my ($package,$vsn,$sfx) = @_;
+    return "${package}_".(stripepoch $vsn).$sfx
+}
+
+sub is_orig_file_of_p_v ($$$) {
+    my ($f, $package, $upstreamvsn) = @_;
+    my $base = source_file_leafname $package, $upstreamvsn, '';
+    return 0 unless $f =~ m/^\Q$base\E\.$orig_f_tail_re$/;
+    return 1;
+}
+
 sub server_branch ($) { return "$branchprefix/$_[0]"; }
 sub server_ref ($) { return "refs/".server_branch($_[0]); }
 
@@ -227,7 +284,7 @@ sub stat_exists ($) {
     my ($f) = @_;
     return 1 if stat $f;
     return 0 if $!==&ENOENT;
-    die "stat $f: $!";
+    confess "stat $f: $!";
 }
 
 sub _us () {
@@ -235,8 +292,8 @@ sub _us () {
 }
 
 sub failmsg {
-    my $s = "error: @_\n";
-    $s =~ s/\n\n$/\n/;
+    my $s = f_ "error: %s\n", "@_";
+    $s =~ s/\n\n$/\n/g;
     my $prefix = _us().": ";
     $s =~ s/^/$prefix/gm;
     return "\n".$s;
@@ -250,12 +307,12 @@ sub ensuredir ($) {
     my ($dir) = @_; # does not create parents
     return if mkdir $dir;
     return if $! == EEXIST;
-    die "mkdir $dir: $!";
+    confess "mkdir $dir: $!";
 }
 
 sub must_getcwd () {
     my $d = getcwd();
-    defined $d or fail "getcwd failed: $!";
+    defined $d or fail f_ "getcwd failed: %s\n", $!;
     return $d;
 }
 
@@ -274,32 +331,32 @@ our @signames = split / /, $Config{sig_name};
 
 sub waitstatusmsg () {
     if (!$?) {
-	return "terminated, reporting successful completion";
+	return __ "terminated, reporting successful completion";
     } elsif (!($? & 255)) {
-	return "failed with error exit status ".WEXITSTATUS($?);
+	return f_ "failed with error exit status %s", WEXITSTATUS($?);
     } elsif (WIFSIGNALED($?)) {
 	my $signum=WTERMSIG($?);
-	return "died due to fatal signal ".
+	return f_ "died due to fatal signal %s",
 	    ($signames[$signum] // "number $signum").
 	    ($? & 128 ? " (core dumped)" : ""); # POSIX(3pm) has no WCOREDUMP
     } else {
-	return "failed with unknown wait status ".$?;
+	return f_ "failed with unknown wait status %s", $?;
     }
 }
 
 sub failedcmd_report_cmd {
     my $intro = shift @_;
-    $intro //= "failed command";
-    { local ($!); printcmd \*STDERR, _us().": $intro:", @_ or die $!; };
+    $intro //= __ "failed command";
+    { local ($!); printcmd \*STDERR, _us().": $intro:", @_ or confess $!; };
 }
 
 sub failedcmd_waitstatus {
     if ($? < 0) {
-	return "failed to fork/exec: $!";
+	return f_ "failed to fork/exec: %s", $!;
     } elsif ($?) {
-	return "subprocess ".waitstatusmsg();
+	return f_ "subprocess %s", waitstatusmsg();
     } else {
-	return "subprocess produced invalid output";
+	return __ "subprocess produced invalid output";
     }
 }
 
@@ -332,12 +389,13 @@ sub shell_cmd {
 
 sub cmdoutput_errok {
     confess Dumper(\@_)." ?" if grep { !defined } @_;
+    local $printdebug_when_debuglevel = $debugcmd_when_debuglevel;
     debugcmd "|",@_;
-    open P, "-|", @_ or die "$_[0] $!";
+    open P, "-|", @_ or confess "$_[0] $!";
     my $d;
     $!=0; $?=0;
     { local $/ = undef; $d = <P>; }
-    die $! if P->error;
+    confess $! if P->error;
     if (!close P) { printdebug "=>!$?\n"; return undef; }
     chomp $d;
     if ($debuglevel > 0) {
@@ -364,7 +422,7 @@ sub link_ltarget ($$) {
     }
     my $r = link $old, $new;
     $r = symlink $old, $new if !$r && $!==EXDEV;
-    $r or die "(sym)link $old $new: $!";
+    $r or fail "(sym)link $old $new: $!\n";
 }
 
 sub hashfile ($) {
@@ -384,6 +442,7 @@ sub git_cat_file ($;$) {
     # in scalar context, just the data
     # if $etype defined, dies unless type is $etype or in @$etype
     our ($gcf_pid, $gcf_i, $gcf_o);
+    local $printdebug_when_debuglevel = $debugcmd_when_debuglevel;
     my $chk = sub {
 	my ($gtype, $data) = @_;
 	if ($etype) {
@@ -396,18 +455,18 @@ sub git_cat_file ($;$) {
     if (!$gcf_pid) {
 	my @cmd = qw(git cat-file --batch);
 	debugcmd "GCF|", @cmd;
-	$gcf_pid = open2 $gcf_o, $gcf_i, @cmd or die $!;
+	$gcf_pid = open2 $gcf_o, $gcf_i, @cmd or confess $!;
     }
     printdebug "GCF>| ", $objname, "\n";
-    print $gcf_i $objname, "\n" or die $!;
+    print $gcf_i $objname, "\n" or confess $!;
     my $x = <$gcf_o>;
     printdebug "GCF<| ", $x;
     if ($x =~ m/ (missing)$/) { return $chk->($1, undef); }
-    my ($type, $size) = $x =~ m/^.* (\w+) (\d+)\n/ or die "$objname ?";
+    my ($type, $size) = $x =~ m/^.* (\w+) (\d+)\n/ or confess "$objname ?";
     my $data;
-    (read $gcf_o, $data, $size) == $size or die "$objname $!";
+    (read $gcf_o, $data, $size) == $size or confess "$objname $!";
     $x = <$gcf_o>;
-    $x eq "\n" or die "$objname ($_) $!";
+    $x eq "\n" or confess "$objname ($_) $!";
     return $chk->($type, $data);
 }
 
@@ -435,22 +494,22 @@ sub git_for_each_ref ($$;$) {
     if (defined $gitdir) {
 	@cmd = ('sh','-ec','cd "$1"; shift; exec "$@"','x', $gitdir, @cmd);
     }
-    open GFER, "-|", @cmd or die $!;
+    open GFER, "-|", @cmd or confess $!;
     debugcmd "|", @cmd;
     while (<GFER>) {
-	chomp or die "$_ ?";
+	chomp or confess "$_ ?";
 	printdebug "|> ", $_, "\n";
-	m#^(\w+)\s+(\w+)\s+(refs/[^/]+/(\S+))$# or die "$_ ?";
+	m#^(\w+)\s+(\w+)\s+(refs/[^/]+/(\S+))$# or confess "$_ ?";
 	$func->($1,$2,$3,$4);
     }
-    $!=0; $?=0; close GFER or die "$pattern $? $!";
+    $!=0; $?=0; close GFER or confess "$pattern $? $!";
 }
 
 sub git_get_ref ($) {
     # => '' if no such ref
     my ($refname) = @_;
     local $_ = $refname;
-    s{^refs/}{[r]efs/} or die "$refname $_ ?";
+    s{^refs/}{[r]efs/} or confess "$refname $_ ?";
     return cmdoutput qw(git for-each-ref --format=%(objectname)), $_;
 }
 
@@ -482,8 +541,8 @@ sub git_check_unmodified () {
 	if ($?==256) {
 	    fail
 		$cached
-		? "git index contains changes (does not match HEAD)"
-		: "working tree is dirty (does not match HEAD)";
+		? __ "git index contains changes (does not match HEAD)"
+		: __ "working tree is dirty (does not match HEAD)";
 	} else {
 	    failedcmd @cmd;
 	}
@@ -533,11 +592,11 @@ sub git_slurp_config_src ($) {
     local $/="\0";
 
     my $r = { };
-    open GITS, "-|", @cmd or die $!;
+    open GITS, "-|", @cmd or confess $!;
     while (<GITS>) {
-	chomp or die;
+	chomp or confess;
 	printdebug "=> ", (messagequote $_), "\n";
-	m/\n/ or die "$_ ?";
+	m/\n/ or confess "$_ ?";
 	push @{ $r->{$`} }, $'; #';
     }
     $!=0; $?=0;
@@ -554,8 +613,8 @@ sub gdr_ffq_prev_branchinfo ($) {
     #    branch         message is undef
     #    weird-symref   } no $symref,
     #    notbranch      }  no $ffq_prev
-    return ('detached', 'detached HEAD') unless defined $symref;
-    return ('weird-symref', 'HEAD symref is not to refs/')
+    return ('detached', __ 'detached HEAD') unless defined $symref;
+    return ('weird-symref', __ 'HEAD symref is not to refs/')
 	unless $symref =~ m{^refs/};
     my $ffq_prev = "refs/$ffq_refprefix/$'";
     my $gdrlast = "refs/$gdrlast_refprefix/$'";
@@ -571,17 +630,19 @@ sub parsecontrolfh ($$;$) {
 	my %opts = ('name' => $desc);
 	$opts{allow_pgp}= $allowsigned || !$dpkgcontrolhash_noissigned;
 	$c = Dpkg::Control::Hash->new(%opts);
-	$c->parse($fh,$desc) or die "parsing of $desc failed";
+	$c->parse($fh,$desc) or fail f_ "parsing of %s failed", $desc;
 	last if $allowsigned;
 	last if $dpkgcontrolhash_noissigned;
 	my $issigned= $c->get_option('is_pgp_signed');
 	if (!defined $issigned) {
 	    $dpkgcontrolhash_noissigned= 1;
-	    seek $fh, 0,0 or die "seek $desc: $!";
+	    seek $fh, 0,0 or confess "seek $desc: $!";
 	} elsif ($issigned) {
-	    fail "control file $desc is (already) PGP-signed. ".
+	    fail f_
+		"control file %s is (already) PGP-signed. ".
 		" Note that dgit push needs to modify the .dsc and then".
-		" do the signature itself";
+		" do the signature itself",
+		$desc;
 	} else {
 	    last;
 	}
@@ -592,9 +653,9 @@ sub parsecontrolfh ($$;$) {
 sub parsecontrol {
     my ($file, $desc, $allowsigned) = @_;
     my $fh = new IO::Handle;
-    open $fh, '<', $file or die "$file: $!";
+    open $fh, '<', $file or fail f_ "open %s (%s): %s", $file, $desc, $!;
     my $c = parsecontrolfh($fh,$desc,$allowsigned);
-    $fh->error and die $!;
+    $fh->error and confess $!;
     close $fh;
     return $c;
 }
@@ -603,7 +664,7 @@ sub parsechangelog {
     my $c = Dpkg::Control::Hash->new(name => 'parsed changelog');
     my $p = new IO::Handle;
     my @cmd = (qw(dpkg-parsechangelog), @_);
-    open $p, '-|', @cmd or die $!;
+    open $p, '-|', @cmd or confess $!;
     $c->parse($p);
     $?=0; $!=0; close $p or failedcmd @cmd;
     return $c;
@@ -613,7 +674,7 @@ sub getfield ($$) {
     my ($dctrl,$field) = @_;
     my $v = $dctrl->{$field};
     return $v if defined $v;
-    fail "missing field $field in ".$dctrl->get_option('name');
+    fail f_ "missing field %s in %s", $field, $dctrl->get_option('name');
 }
 
 sub parsechangelog_loop ($$$) {
@@ -621,22 +682,105 @@ sub parsechangelog_loop ($$$) {
     # @$clogcmd is qw(dpkg-parsechangelog ...some...options...)
     # calls $fn->($thisstanza, $desc);
     debugcmd "|",@$clogcmd;
-    open CLOGS, "-|", @$clogcmd or die $!;
+    open CLOGS, "-|", @$clogcmd or confess $!;
     for (;;) {
 	my $stanzatext = do { local $/=""; <CLOGS>; };
 	printdebug "clogp stanza ".Dumper($stanzatext) if $debuglevel>1;
 	last if !defined $stanzatext;
 
 	my $desc = "$descbase, entry no.$.";
-	open my $stanzafh, "<", \$stanzatext or die;
+	open my $stanzafh, "<", \$stanzatext or confess;
 	my $thisstanza = parsecontrolfh $stanzafh, $desc, 1;
 
 	$fn->($thisstanza, $desc);
     }
-    die $! if CLOGS->error;
+    confess $! if CLOGS->error;
     close CLOGS or $?==SIGPIPE or failedcmd @$clogcmd;
 }	
 
+sub make_commit_text ($) {
+    my ($text) = @_;
+    my ($out, $in);
+    my @cmd = (qw(git hash-object -w -t commit --stdin));
+    debugcmd "|",@cmd;
+    print Dumper($text) if $debuglevel > 1;
+    my $child = open2($out, $in, @cmd) or confess $!;
+    my $h;
+    eval {
+	print $in $text or confess $!;
+	close $in or confess $!;
+	$h = <$out>;
+	$h =~ m/^\w+$/ or confess;
+	$h = $&;
+	printdebug "=> $h\n";
+    };
+    close $out;
+    waitpid $child, 0 == $child or confess "$child $!";
+    $? and failedcmd @cmd;
+    return $h;
+}
+
+sub reflog_cache_insert ($$$) {
+    my ($ref, $cachekey, $value) = @_;
+    # you must call this in $maindir
+    # you must have called record_maindir
+
+    # When we no longer need to support squeeze, use --create-reflog
+    # instead of this:
+    my $parent = $ref; $parent =~ s{/[^/]+$}{};
+    ensuredir "$maindir_gitcommon/logs/$parent";
+    my $makelogfh = new IO::File "$maindir_gitcommon/logs/$ref", '>>'
+      or confess $!;
+
+    my $oldcache = git_get_ref $ref;
+
+    if ($oldcache eq $value) {
+	my $tree = cmdoutput qw(git rev-parse), "$value:";
+	# git update-ref doesn't always update, in this case.  *sigh*
+	my $authline = (ucfirst _us()).
+	    ' <'._us().'@example.com> 1000000000 +0000';
+	my $dummy = make_commit_text <<ENDU.(__ <<END);
+tree $tree
+parent $value
+author $authline
+committer $authline
+
+ENDU
+Dummy commit - do not use
+END
+	runcmd qw(git update-ref -m), _us()." - dummy", $ref, $dummy;
+    }
+    runcmd qw(git update-ref -m), $cachekey, $ref, $value;
+}
+
+sub reflog_cache_lookup ($$) {
+    my ($ref, $cachekey) = @_;
+    # you may call this in $maindir or in a playtree
+    # you must have called record_maindir
+    my @cmd = (qw(git log -g), '--pretty=format:%H %gs', $ref);
+    debugcmd "|(probably)",@cmd;
+    my $child = open GC, "-|";  defined $child or confess $!;
+    if (!$child) {
+	chdir $maindir or confess $!;
+	if (!stat "$maindir_gitcommon/logs/$ref") {
+	    $! == ENOENT or confess $!;
+	    printdebug ">(no reflog)\n";
+	    finish 0;
+	}
+	exec @cmd; die f_ "exec %s: %s\n", $cmd[0], $!;
+    }
+    while (<GC>) {
+	chomp;
+	printdebug ">| ", $_, "\n" if $debuglevel > 1;
+	next unless m/^(\w+) (\S.*\S)$/ && $2 eq $cachekey;
+	close GC;
+	return $1;
+    }
+    confess $! if GC->error;
+    failedcmd unless close GC;
+    return undef;
+}
+
 # ========== playground handling ==========
 
 # terminology:
@@ -685,18 +829,13 @@ sub parsechangelog_loop ($$$) {
 
 # ----- maindir -----
 
-# these three all go together
-our $maindir;
-our $maindir_gitdir;
-our $maindir_gitcommon;
-
 our $local_git_cfg;
 
 sub record_maindir () {
     if (!defined $maindir) {
 	$maindir = must_getcwd();
 	if (!stat "$maindir/.git") {
-	    fail "cannot stat $maindir/.git: $!";
+	    fail f_ "cannot stat %s/.git: %s", $maindir, $!;
 	}
 	if (-d _) {
 	    # we fall back to this in case we have a pre-worktree
@@ -718,15 +857,16 @@ sub ensure_a_playground_parent ($) {
     record_maindir();
     $spc = "$maindir_gitdir/$spc";
     my $parent = dirname $spc;
-    mkdir $parent or $!==EEXIST
-	or fail "failed to mkdir playground parent $parent: $!";
+    mkdir $parent or $!==EEXIST or fail f_
+	"failed to mkdir playground parent %s: %s", $parent, $!;
     return $spc;
 }    
 
 sub ensure_a_playground ($) {
     my ($spc) = @_;
     $spc = ensure_a_playground_parent $spc;
-    mkdir $spc or $!==EEXIST or fail "failed to mkdir a playground $spc: $!";
+    mkdir $spc or $!==EEXIST or fail f_
+	"failed to mkdir a playground %s: %s", $spc, $!;
     return $spc;
 }    
 
@@ -734,7 +874,8 @@ sub fresh_playground ($) {
     my ($spc) = @_;
     $spc = ensure_a_playground_parent $spc;
     rmtree $spc;
-    mkdir $spc or fail "failed to mkdir the playground $spc: $!";
+    mkdir $spc or fail f_
+	"failed to mkdir the playground %s: %s", $spc, $!;
     return $spc;
 }
 
@@ -761,11 +902,11 @@ sub playtree_setup (;$) {
     #   $maindir_gitdir     contains our main working "dgit", HEAD, etc.
     #   $maindir_gitcommon  the shared stuff, including .objects
     rmtree('.git/objects');
-    symlink "$maindir_gitcommon/objects",'.git/objects' or die $!;
+    symlink "$maindir_gitcommon/objects",'.git/objects' or confess $!;
     ensuredir '.git/info';
-    open GA, "> .git/info/attributes" or die $!;
-    print GA "* $negate_harmful_gitattrs\n" or die $!;
-    close GA or die $!;
+    open GA, "> .git/info/attributes" or confess $!;
+    print GA "* $negate_harmful_gitattrs\n" or confess $!;
+    close GA or confess $!;
 }
 
 1;
